import * as Bluebird from 'bluebird';

function defer(): any {
    let resolve: any;
    let reject: any;
    let promise: any = new Bluebird(function(): any {
        resolve = arguments[0];
        reject = arguments[1];
    });
    return {
        resolve: resolve,
        reject: reject,
        promise: promise
    };
};

const CAMEL_REGEXP: RegExp = /[A-Z]/g;

const camelize = (camelCase: any) => camelCase.replace(CAMEL_REGEXP, (x: any, pos: any) => (pos > 0 ? '_' : '') + x.toLowerCase());

const SLICE_ = Array.prototype.slice;

function removeFromArray(array: Array<any>, object: Object): Boolean | null {
    if (!array) {
        return null;
    }

    const index = array.indexOf(object);
    const exists = index !== -1;

    if (exists) {
        array.splice(index, 1);
    }
    return exists;
};

class EventEmitter {

    public _events: any = {};
    public _any: Array<any> = [];

    _getListeners (eventName: string, createIfMissing: boolean): any {
        const events = this._getEvents();
        let listeners = events[eventName];
        if (createIfMissing && !listeners) {
            listeners = events[eventName] = [];
        }

        return listeners;
    }

    _getEvents (): any {
        let events: any = this._events;
        if (!events) {
            events = this._events = {};
        }

        return events;
    }

    _getAnyListeners (createIfMissing: boolean): any {
        let any = this._any;
        if (createIfMissing && !any) {
            any = this._any = [];
        }

        return any;
    }

    bindAll (object: any, context: any): any {

        context = context || this;

        const bindMethod = function(method: any): any  {
            context.addListener(camelize(method.substr(2)), function(): any {
                const args = SLICE_.call(arguments, 0);
                args.push(context);
                object[method].apply(object, args);
            });
        };

        for (let method in object) {
            if (typeof object[method] === 'function' && method.substr(0, 2) === 'on') {
                bindMethod(method);
            }
        }
    }

    emitAsync (event: any): any {

        let pending = this.listeners(event).length;
        let deferred = defer();
        let done = () => {

            if (!--pending) {
                deferred.resolve();
            }
        };

        this.emit(event, done);

        if (!pending) {
            deferred.resolve();
        }

        return deferred.promise;
    }

    emit (eventName: any, ...args: Array<any>): any {
        const emitListeners = (listeners: any, type: any, arg: any) => {

            if (!listeners) {
                return;
            }

            listeners
                .slice()
                .forEach((listener: any, index: any) => {
                    if (listener._once) {
                        if (listener._fired) {
                            return;
                        }
                        listeners.splice(index, 1);
                        listener._fired = true;
                    }
                    if (!!listener.emit) {
                        listener.emit(type, arg);
                    } else {
                        listener.apply(this, arg);
                    }
                });
        };

        const listeners = this._getListeners(eventName, false);

        // If there is no 'error' event listener then throw.
        if (eventName === 'error' && !listeners) {
            const error = arguments[1];
            if (error instanceof Error) {
                throw error; // Unhandled 'error' event
            } else {
                // At least give some kind of context to the user
                const err = new Error('Uncaught, unspecified "error" event. (' + error + ')');
                ///err.context = error;
                throw err;
            }
        }
        emitListeners(listeners, eventName, args);
        emitListeners(this._any, eventName, [eventName, ...args]);
    }

    on (eventName: any, listener: any): any {

        if (typeof listener !== 'function') {
            throw new TypeError('"listener" argument must be a function');
        }

        // To avoid recursion in the case that eventName === "newListener"! Before
        // adding it to the listeners, first emit "newListener".
        if (this._getListeners('newListener', false)) {
            this.emit('newListener', eventName, listener);
        }

        const listeners = this._getListeners(eventName, true);
        listeners.push(listener);
        return this;
    }

    once (eventName: any, listener: any): any {

        if (typeof listener !== 'function') {
            throw new TypeError('"listener" argument must be a function');
        }

        listener._once = true;
        this.on(eventName, listener);
    }

    onAny (listener: any): any {
        this._getAnyListeners(true).push(listener);
        return this;
    }

    offAny (listener: any): any {
        removeFromArray(this._getAnyListeners(false), listener);
        return this;
    }

    // emits a 'removeListener' event if the listener was removed
    off (eventName: any, listener: any): any {

        if (typeof listener !== 'function') {
            throw new TypeError('"listener" argument must be a function');
        }

        let list: any = this._getListeners(eventName, false);

        // Nothing to do here...
        if (!list) {
            return this;
        }

        const removed = removeFromArray(list, listener);

        if (removed && this._getEvents().removeListener) {
            this.emit('removeListener', eventName, listener);
        }

        return this;
    }

    listeners (eventName: any): any {
        const listeners = this._getListeners(eventName, false);
        return listeners ? listeners.slice() : [];
    }

    removeAllListeners (eventName: any): any {

        const events = this._getEvents();

        // nothing to do here...
        if (!events) {
            return this;
        }

        if (!eventName) {
            Object.keys(events).forEach((type: any): any => this.removeAllListeners(type));
            return this;
        } else {
            const listeners = this._getListeners(eventName, false);
            if (!listeners) {
                return;
            }
            listeners
                .slice()
                .forEach((listener: any): any => this.off(eventName, listener));
            delete events[eventName];
        }
        return this;
    }
    addListener(eventName: any, listener: any): any {}
    removeListener(eventName: any, listener: any): any {}
    listenerCount (eventName: any): any {
        return this.listeners(eventName).length;
    }

    static listenerCount (emitter: any, eventName: any): any {
        return emitter.listenerCount(eventName);
    }
}

EventEmitter.prototype.addListener = EventEmitter.prototype.on;
EventEmitter.prototype.removeListener = EventEmitter.prototype.off;

export default EventEmitter;
